#include <mpi.h> 
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#define CTE 2.506631   /*CTE=sqrt(2.0*PI)*/
#define WORK 0
#define DONE 1
#define RESULT 2

 double ** test(double **data,double **X,double *w,int min,int r,int n_trai,int p){
   double **temp;
   int i;
   for(i=0;i<n_trai;i++){
     w[i]=1.0/(double)n_trai;
     if(i<min)
       data[i]=X[i];
     else
       data[i]=X[i+r];
   }
   
   temp=calloc(r,sizeof(double *)); 
   for(i=0;i<r;i++)
     temp[i]= X[i+min]; 
   return temp;
 }

 double ** test_par(double **data,double **X,double *w,int min,int r,int n_trai,int p,int rank,int size){
   int residuo_trai=n_trai-size*(n_trai/size),q_trai=n_trai/size;
   double **temp;
   int i;
if(rank<residuo_trai)
  q_trai=q_trai+1;

 for(i=0;i<q_trai;i++)
  w[i]=1.0/n_trai;   


   for(i=0;i<n_trai;i++){
   
     if(i<min)
       data[i]=X[i];
     else
       data[i]=X[i+r];
   }
   
   temp=calloc(r,sizeof(double *)); 
   for(i=0;i<r;i++)
     temp[i]= X[i+min]; 
   return temp;
 }


double ** training(double **X,int n_trai,int m,int p,double *tab){
  double **temp;
  double *class=NULL;
  double min=0.0;
  long r;
  int i;
  class=calloc(m+1,sizeof(double));
  for(i=0;i<m+1;i++)
    class[i]=0.0;
  temp=calloc(n_trai,sizeof(double *));
  for(i=0;i<n_trai;i++){
    r=rand();
    temp[i]= X[r%n_trai];
    class[(int)temp[i][p-1]]+=1.0;
    
  }  
  min=class[1];
  for(i=2;i<m+1;i++){
    if(class[i]<=min)
      min=class[i];
  }
  *tab=min;
  return temp;				
}
 
double ** training_par(double **X,int n_trai,int m,int p,int rank){
  long r;
  int i,critico=5;              
  int *indice;
  double **temp;
  double *class=NULL;
  double tabulate=0.0;
	
  temp=calloc(n_trai,sizeof(double *));
  indice=calloc(n_trai,sizeof(int));

  if(rank==0){

    class=calloc(m+1,sizeof(double));
    for(i=0;i<m+1;i++)
      class[i]=0.0;
 
    tabulate=0.0;
    while(tabulate<=(double)critico){    
      for(i=0;i<n_trai;i++){
	r=rand();
	indice[i]=r%n_trai;
	temp[i]= X[indice[i]];
	class[(int)temp[i][p-1]]+=1.0;
      }  
      tabulate=class[1];
      for(i=2;i<m+1;i++){
	if(class[i]<=tabulate)
	  tabulate=class[i];
      }
    }  
    MPI_Bcast(indice,n_trai,MPI_INT,0,MPI_COMM_WORLD);
}
  
  else{
    MPI_Bcast(indice,n_trai,MPI_INT,0,MPI_COMM_WORLD);
    for(i=0;i<n_trai;i++)
      temp[i]= X[indice[i]];
               
  }
  
  return temp;				
}

double **trai_weight(double **data,double *w,int min,int max,int n_trai){
  int i,j,p,r,num=n_trai*10;
  int *indice=NULL;
  double s=0.0,t=0.0;
  double *prob_acum=NULL;
  double **temp;      
  indice = calloc(n_trai,sizeof(int));
  temp=calloc(n_trai,sizeof(double *));
  prob_acum = calloc(n_trai,sizeof(double));
  for(i=0;i<n_trai;i++)
     s=s+w[i];
  for(i=0;i<n_trai;i++){
    w[i]=w[i]/s;
    t+=w[i];
    prob_acum[i]=t;
  }      
  for(i=0;i<n_trai;i++){
    p=rand()%num;
    for(j=0;j<n_trai;j++){ 
      if(((double)p/(double)num)<prob_acum[j]){
	indice[i]=j;
	break;
      }
    }
    temp[i]=data[indice[i]];
  }
  return temp;
}


double **trai_weight_par(double **data,double *w,int min,int max,int n_trai,int rank,int size){
  int i,j,p,count,l,num=n_trai*10;
  int residuo_trai=n_trai-size*(n_trai/size),q_trai=n_trai/size;
  int *indice=NULL;
  double s=0.0,t=0.0,result;
  double *prob_acum=NULL;
  double *w_temp=NULL;
  double **temp;
  MPI_Status status;
      
  indice = calloc(n_trai,sizeof(int));
  temp=calloc(n_trai,sizeof(double *));

  if(rank<residuo_trai)
    q_trai=q_trai+1;

  for(i=0;i<q_trai;i++)
          s=s+w[i];
   
  MPI_Allreduce(&s,&result,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
   
   for(i=0;i<q_trai;i++)
        w[i]=w[i]/result;
  
   if(rank==0){

             prob_acum = calloc(n_trai,sizeof(double));
	     w_temp = calloc(q_trai+1,sizeof(double));

	   if(rank<residuo_trai)
	      count=q_trai+1;
           else
	       count=q_trai;

	   l=count; 
  	   for(i=0;i<size-1;i++){
 
	   if(i+1<residuo_trai)
 	       count=q_trai+1;
           else
	       count=q_trai;
	
	    MPI_Recv(w_temp,count,MPI_DOUBLE,i+1,0,MPI_COMM_WORLD,&status);
	    

	    for(j=0;j<count;j++){
	            *(w+l)=*(w_temp+j);
	            l++;
	      }
          }

	 for(i=0;i<n_trai;i++){
	   t+=w[i];
	   prob_acum[i]=t;
	 }


	 for(i=0;i<n_trai;i++){
	   p=rand()%num;
	   for(j=0;j<n_trai;j++){ 
	     if(((double)p/(double)num)<prob_acum[j]){
	       indice[i]=j;
	       break;
	     }
	   }
	   temp[i]=data[indice[i]];
	 }
	 MPI_Bcast(indice,n_trai,MPI_INT,0,MPI_COMM_WORLD);
   }
   else{
     MPI_Ssend(w,q_trai+1,MPI_DOUBLE,0,0,MPI_COMM_WORLD);
     MPI_Bcast(indice,n_trai,MPI_INT,0,MPI_COMM_WORLD);
     for(i=0;i<n_trai;i++)
       temp[i]=data[indice[i]];
   }
   return temp;
}

int * randperm(int n ){
       int i,j,min;
       int * a=NULL;
       int * indice=NULL;
       indice=(int *)calloc(n,sizeof(int));
       a=(int *)calloc(n,sizeof(int));   
       srand(time(NULL)); 
       for(i=0;i<n;i++)
           a[i]=rand()%n; 
       for ( i = 0; i < n;i++ ){
           min=i;
           for (j = 0; j < n; j++ )
                if ( a[j] < a[min] ) {
                    min=j;
                }
           a[min]=n;
           indice[i]=min; 
       }
       return indice;
       }

/*funcion que calcula el kernel gaussiano*/
double kernel(double x){
       return 1.0/(exp((x*x)/2.0))*CTE;        /*CTE =sqrt(2.0*PI)*/
       }

double * banda(double **X, int n,int p){
       double s,d, Xmean,*h=NULL;
       int i,v;
       h=calloc(n,sizeof(double)); 
       for(v=0;v<p ;v++){
            Xmean=0;
            d=0;
            s=0;
            for(i=0;i<n;i++){
                 s+=X[i][v];
            }
            Xmean =s/n;       
            for(i=0;i<n;i++){
                d+=(X[i][v]-Xmean)*(X[i][v]-Xmean);
            }
            d=sqrt(d/(n-1));    
                                
            h[v]=d*(pow(4.0,1.0/(double)(p+4))/pow((double)((p+2)*n),1.0/(double)(p+4)));  
       }
       return h;
}
int error(int **M,int * k_indices, double *beta,int n_test,int n_boos,int m,int min ){
  int i,j,e=0,k;
  double max;
  double *clase=NULL;
  clase=calloc(m+1,sizeof(double)); 
  for(j=0;j<n_test;j++){
    for(i=0;i<m+1;i++) clase[i]=0.0;  
    for(i=0;i<n_boos;i++) {clase[M[i][j]]+=log(1.0/beta[i]);}
    max=0.0;
    k=0;   
    for(i=0;i<m+1;i++){
      if(clase[i]>max){   
	max=clase[i];
	k=i;      
      }
    } 
    if(k_indices[j+min]!=k) 
      e++;
  }
  return e;
}

int error_par(int **M,int *k_indices ,double *beta,int n_boos,int n_cross,int m,int min ,int rank,int size){
  int i,j,e=0,k,q=n_cross/size,residuo=n_cross-size*(n_cross/size);
  double max;
  double *clase=NULL;
  clase=calloc(m+1,sizeof(double)); 
  if(rank<residuo){
    for(j=0;j<q+1;j++){
      for(i=0;i<m+1;i++) clase[i]=0.0;  
      for(i=0;i<n_boos;i++) {clase[M[i][j]]+=log(1.0/beta[i]);}
      max=0.0;
      k=0;   
      for(i=0;i<m+1;i++){
	if(clase[i]>max){   
	  max=clase[i];
	  k=i;      
	}
      } 
      if(k_indices[j+min+rank*(q+1)]!=k) 
	e++;
    }

  }
  else {
    for(j=0;j<q;j++){
      for(i=0;i<m+1;i++) clase[i]=0.0;  
      for(i=0;i<n_boos;i++) {clase[M[i][j]]+=log(1.0/beta[i]);}
      max=0.0;
      k=0;   
      for(i=0;i<m+1;i++){
	if(clase[i]>max){   
	  max=clase[i];
	  k=i;      
	}
	
      } 
      if(k_indices[j+min+residuo+rank*q]!=k) 
	e++;
    }
  }
  return e;
}

int  *maximos(double **f,int n,int m ){
      int i,j,k;
      int *max=NULL;
      double maximo;
      max = calloc(n,sizeof(int)); 
      for(j=0;j<n;j++){
             maximo=0;
             for(i=0;i<m;i++){
                   if(f[i][j]>=maximo){
                        maximo=f[i][j];
                        k=i+1;
                    }
             }
             max[j]=k;     
       }
       return max;
       }


double geometrica(double *f,int n_trai){
  int i;
  double g=0.0;
  for(i=0;i<n_trai;i++){
    g=g+log(f[i]);
  }
  g=exp(g/(double)n_trai);
  return g;
}

double geometrica_par(double *f,int q,int n_trai){
     int i;
     double g=0.0;
     for(i=0;i<q;i++){
       g=g+log(f[i]);
     }
     g=exp(g/(double)n_trai);
     return g;
}

double * fpiloto(double **x,double *h,double **X,double *g, int p,int m_elem,int n_cross,int N){
  int v,j,i;
  double s=0.0,t,k,H=1.0,alpha=-0.5;
  double *f=NULL;
  double pro=(double)m_elem/(double)(N-n_cross);
  double geometrica(double *,int);
  f=calloc(m_elem,sizeof(double));
  for(v=0;v<p;v++)
    H*=h[v];
  H=H*(double)m_elem;
  for (i=0;i<m_elem;i++){
    s=0.0;                   
    for(j=0;j<m_elem ;j++){
      k=1.0;
      for(v=0;v<p;v++){
	t=kernel((x[i][v]-X[j][v])/h[v]);
	k=k*t;
      }
      s=s+k;
    }
    f[i]=(s/H);
  }
  *g=geometrica(f,m_elem);
  for(i=0;i<m_elem;i++)
    f[i]= pow(f[i]/(*g),alpha);
  return f;
}

double * fpiloto_par(double **x,double *h,double **X,double *g, int p,int m_elem,int n_cross,int N,int rank,int size){
  MPI_Status status;
  int v,j,l=0,i,count,residuo=m_elem-size*(m_elem/size),q=m_elem/size;
  double s=0.0,result,t,k,H=1.0,alpha=-0.5;
  double *f=NULL;
  double *fhat=NULL;
  double *temp=NULL;
  double pro=(double)m_elem/(double)(N-n_cross);
  double geometrica(double *,int);
  temp=calloc(q+1,sizeof(double));
  fhat=calloc(m_elem,sizeof(double));
  for(v=0;v<p;v++)
    H*=h[v];  
  H=H*(double)m_elem;
  if(rank<residuo){
    f=calloc(q+1,sizeof(double));/*se asigna memoria dinamica para el vector de salida*/
    for (i=rank*(q+1);i<(rank+1)*(q+1);i++){
      s=0.0;                   
      for(j=0;j<m_elem ;j++){
	k=1.0;
	for(v=0;v<p;v++){
	  t=kernel((x[i][v]-X[j][v])/h[v]);
	  k=k*t;
	}
	s=s+k;                         /*para acumular las productorias y asi obtener la sumatoria.*/
      }
      f[i-rank*(q+1)]=(s/H);   /*se almacena los estimados para cada fila del test*/
    }
    *g=geometrica_par(f,q+1,m_elem);
  }
  else {
    f=calloc(q,sizeof(double));
    for (i=residuo+rank*q;i<residuo+(rank+1)*q;i++){
      s=0.0;                   
      for(j=0;j<m_elem ;j++){
	k=1.0;
	for(v=0;v<p;v++){
	  t=kernel((x[i][v]-X[j][v])/h[v]);
	  k=k*t;
	}
	s=s+k;                         /*para acumular las productorias y asi obtener la sumatoria.*/
      }
      f[i-(residuo+rank*q)]=(s/H);   /*se almacena los estimados para cada fila del test*/
    }
    *g=geometrica_par(f,q,m_elem);
  }
  MPI_Reduce(g,&result,1,MPI_DOUBLE,MPI_PROD,0,MPI_COMM_WORLD);
  *g=result;
  if(rank==0){
    if(rank<residuo)
      count=q+1;
    else
      count=q;
    for(j=0;j<count;j++){
      *(fhat+l)=*(f+j);
      l++;
    }
    for(i=0;i<size-1;i++){
      if(i+1<residuo)
	count=q+1;
      else
	count=q;
      MPI_Recv(temp,count,MPI_DOUBLE,i+1,0,MPI_COMM_WORLD,&status);
      for(j=0;j<count;j++){
	*(fhat+l)=*(temp+j);
	l++;
      }
    }
    for(i=0;i<m_elem;i++)
      fhat[i]= pow(fhat[i]/result,alpha);
  }
  else{
    if(rank<residuo)
      MPI_Ssend(f,q+1,MPI_DOUBLE,0,0,MPI_COMM_WORLD);
    else
      MPI_Ssend(f,q,MPI_DOUBLE,0,0,MPI_COMM_WORLD);
  }
  MPI_Bcast(fhat,m_elem,MPI_DOUBLE,0,MPI_COMM_WORLD);
  return fhat;
}

double * estimado(double **x,double *h,double **X,int n_cross, int p,int m_elem,int N){
  int v,j,i;
  double s=0.0,t,k,g,H=0.0,alpha =-0.5,lamda;
  double *fhat=NULL;
  double *f=NULL;
  double pro=(double)m_elem/(double)(N-n_cross);
  double geometrica(double *,int);
  
  fhat = calloc(m_elem,sizeof(double)); 
  f=calloc(n_cross,sizeof(double));
  for(v=0;v<p;v++)
    H+=log(h[v]);
  H=exp(H)*(double)m_elem;
  
  fhat = fpiloto(X,h,X,&g,p,m_elem,n_cross,N);
  
  for (i=0;i<n_cross;i++){
    s=0.0;
    for(j=0;j<m_elem ;j++){
      k=1.0;
      for(v=0;v<p;v++){
	t=kernel((x[i][v]-X[j][v])/(fhat[j]*h[v]));
	k=k*t;
      }
      s=s+(k/pow(fhat[j],p-1)); 
    }
    f[i]=pro*(s/H);
  }
  
  return f;
}

double * adap_estimado_par(double **x,double *h,double **X,int n_cross, int p,int m_elem,int N,int rank, int size){
  int v,j,i,residuo=n_cross-size*(n_cross/size),q=n_cross/size,mq=m_elem/size;
  double s=0.0,t,k,g,result,H=0.0,alpha =-0.5,lamda;
  double *fhat=NULL;
  double *fhatp=NULL;
  double *f=NULL;
  double pro=(double)m_elem/(double)(N-n_cross);
  fhat = calloc(m_elem,sizeof(double)); 
  for(v=0;v<p;v++)
    H+=log(h[v]); 
  H=exp(H)*(double)m_elem;
  fhat = fpiloto_par(X,h,X,&g,p,m_elem,n_cross,N,rank,size);
  if(rank<residuo){
    f=calloc(q+1,sizeof(double));/*se asigna memoria dinamica para el vector de salida*/
    for (i=rank*(q+1);i<(rank+1)*(q+1);i++){
      s=0.0;                   
      for(j=0;j<m_elem ;j++){
	k=1.0;
	for(v=0;v<p;v++){
	  t=kernel((x[i][v]-X[j][v])/(fhat[j]*h[v]));
	  k=k*t;
	}
	s=s+(k/pow(fhat[j],p-1));                         /*para acumular las productorias y asi obtener la sumatoria.*/
      }
      f[i-rank*(q+1)]=pro*(s/H);   /*se almacena los estimados para cada fila del test*/
    }
    
  }
  else {
    f=calloc(q,sizeof(double));
    for (i=residuo+rank*q;i<residuo+(rank+1)*q;i++){
      s=0.0;                   
      for(j=0;j<m_elem ;j++){
	k=1.0;
	for(v=0;v<p;v++){
	  t=kernel((x[i][v]-X[j][v])/(fhat[j]*h[v]));
	  k=k*t;
	}
	s=s+(k/pow(fhat[j],p-1));                         /*para acumular las productorias y asi obtener la sumatoria.*/
      }
      f[i-(residuo+rank*q)]=pro*(s/H);   /*se almacena los estimados para cada fila del test*/
    }
  }
  return f;
}

double **estimated(double ** x, double ** X,int n_trai, int p,int m,int r,int N){
  double **f;
  double ** temp;
  double *h=NULL;
  int i,j,cont;
  
  f=calloc(m,sizeof(double *));
  temp=calloc(n_trai,sizeof(double *)); 
  h=calloc(p,sizeof(double));
  for(j=0;j<m;j++){
    cont=0;
    for(i=0;i<n_trai;i++)
      if(X[i][p-1]==(double)(j+1)){
	temp[cont++]= X[i];
      }
    
    h=banda(temp,cont,p-1);       
    f[j] = estimado(x,h,temp,r,p-1,cont,N);
  }
  return f;
}

double **estimated_par(double ** x, double ** X,int n, int p,int m,int r,int N,int rank,int size){
  double **f;
  double ** temp;
  double *h=NULL;
  int i,j,cont;
  
  f=calloc(m,sizeof(double *)); 
  temp=calloc(n,sizeof(double *));
  h=calloc(p,sizeof(double));   
  for(j=0;j<m;j++){
    cont=0;
    for(i=0;i<n;i++)
      if(X[i][p-1]==(double)(j+1)){
	temp[cont++]= X[i];
      }
    
    h=banda(temp,cont,p-1);       
    /* f[j] = estimado(x,h,temp,r,p-1,cont,N,rank,size)*/;
    f[j]=adap_estimado_par(x,h,temp,r,p-1,cont,N,rank,size);
  }
      return f;
}

double max_error(double **ftrai,double **f,double *w,int *k_indices,int *max,int *k,int r,int min,int m,int n_trai ){
  int i,j,cont;
  double e=0.0,maximo;
  
  for(j=0;j<r;j++){
    maximo=0.0; 
    for(i=0;i<m;i++){
      if(f[i][j]>=maximo){
	maximo=f[i][j];
	cont=i+1;
	
      }
    }
    max[j]=cont;
  }
  
  for(j=0;j<n_trai;j++){
    maximo=0.0; 
    for(i=0;i<m;i++){
      if(ftrai[i][j]>=maximo){
	maximo=ftrai[i][j];
	cont=i+1;
      }
    }
    
    if(j<min){
      if(k_indices[j]!=cont){
	e+=w[j];
	k[j]=0;
	
      }
      else
	k[j]=1;
    }
    else{	   
      if(k_indices[j+r]!=cont){
	e+=w[j];
	k[j]=0;
	
      }
      else
	k[j]=1;    
    }
  }
  
  return e;
}

double max_error_par(double **ftrai,double **f,double *w,int *k_indices,int *max,int *k,int r,int min,int m,int n_trai,int rank,int size ){
      int i,j,cont,residuo_r=r-size*(r/size) ,q_r=r/size;
      int residuo_trai=n_trai-size*(n_trai/size),q_trai=n_trai/size;
      double e=0.0,maximo,result;
      
      if(rank<residuo_r)
	q_r=q_r+1;

      for(j=0;j<q_r;j++){
             maximo=0.0; 
             for(i=0;i<m;i++){
                   if(f[i][j]>=maximo){
                        maximo=f[i][j];
                        cont=i+1;
                    }
             }
	     max[j]=cont;
	    
       }

      if(rank<residuo_trai){
          for(j=0;j<q_trai+1;j++){
             maximo=0.0;
             for(i=0;i<m;i++){
                   if(ftrai[i][j]>=maximo){
                        maximo=ftrai[i][j];
                        cont=i+1;

                    }
             }

	     if(j+rank*(q_trai+1)<min){
                    if(k_indices[j+rank*(q_trai+1)]!=cont){ 
                          e+=w[j];
		           k[j]=0;

		    }
		    else
		        k[j]=1;

	     }
	     else{
	             if(k_indices[j+rank*(q_trai+1)+r]!=cont){ 
                          e+=w[j];
		           k[j]=0;
			   

		     }
		     else
		         k[j]=1;


	     }
	  }

      }

      else {
             for(j=0;j<q_trai;j++){
	       maximo=0.0;
                for(i=0;i<m;i++){
                   if(ftrai[i][j]>=maximo){
                        maximo=ftrai[i][j];
                        cont=i+1;
                    }
                 }
		
		if(j+rank*q_trai+residuo_trai<min){
                    if(k_indices[j+rank*q_trai+residuo_trai]!=cont){ 
                          e+=w[j];
		           k[j]=0;
			   

		    }
		    else
		      k[j]=1;

		}
		else{
	             if(k_indices[j+rank*q_trai+residuo_trai+r]!=cont){ 
                          e+=w[j];
		           k[j]=0;
			  
		     }
		     else
		       k[j]=1;
		}

	     }
     
      }
        
      return e;
}


void normalizador(double *w,int n_trai){
  int j;
  double s=0.0;
      for(j=0;j<n_trai;j++){
	s+=w[j];
      }

      for(j=0;j<n_trai;j++){
	w[j]=w[j]/(double)s;

      }
}

int *max_estimado(double **data,double ** x, double ** X,int *k_indices,double *w,int n, int p,int m,int r,int min,double *eb){
       
  int i,j,tol=25,cont,n_trai= n-r,max=min+r,critico=5;
  int *maximos=NULL;
  int *k=NULL;
  double maximo,e=0.0,tabulate; 
  double **f;
  double **ftrai;
  
  k=calloc(n_trai,sizeof(int));
  f=calloc(m,sizeof(double *));   
  ftrai=calloc(m,sizeof(double *));   
  maximos = calloc(r,sizeof(int));
  
  
  f=estimated( x,X,n_trai,p,m,r,n);
				   
  ftrai=estimated(data,X,n_trai,p,m,n_trai,n);
  e= max_error(ftrai,f,w,k_indices,maximos,k,r,min, m,n_trai );
  
  if((e<0.5)&&(e!=0.0)){     
    for(j=0;j<n_trai;j++){
      if(k[j]==1)
	w[j]=(w[j])/(2*e);
      else
	w[j]=(w[j])/(2.0*(1.0-e));
    }
  }
  else{ 
    for(j=0;j<tol;j++){
      tabulate=0.0;
      while(tabulate<=(double)critico){
	X=training(data,n_trai,m,p,&tabulate);
      }
      f=estimated( x,X,n_trai,p,m,r,n); 
      ftrai=estimated(data,X,n_trai,p,m,n_trai,n); 
      for(i=0;i<n_trai;i++)
	w[i]=1.0/(double)n_trai;
      e= max_error(ftrai,f,w,k_indices,maximos,k,r,min, m ,n_trai);
      if((e<0.5)&&(e!=0)){
	
	break;
      }
    }
    
    for(j=0;j<n_trai;j++){
      if(k[j]==1)
	w[j]=(w[j]*(double)n_trai)/(2*e);
      else
	w[j]=(w[j]*(double)n_trai)/(2*((double)n_trai-e));
    }
  }
  
  *eb=e;
  return maximos;
  
}

int *max_estimado_par(double **data,double ** x, double ** X,int *k_indices,double *w,int n, int p,int m,int r,int min,double *eb,int rank,int size){
       
       int i,j,tol=25,n_trai= n-r,critico=3;
       int residuo_trai=n_trai-size*(n_trai/size),q_trai=n_trai/size;
       int *maximos=NULL;
       int *k=NULL;
       double e=0.0,par_e,tabulate; 
       double **f;
       double **ftrai;
 
       k=calloc(n_trai,sizeof(int));
       f=calloc(m,sizeof(double *));   
       ftrai=calloc(m,sizeof(double *));   
       maximos = calloc(r,sizeof(int));
       f=estimated_par( x,X,n_trai,p,m,r,n,rank,size);
       ftrai=estimated_par(data,X,n_trai,p,m,n_trai,n,rank,size);
       par_e= max_error_par(ftrai,f,w,k_indices,maximos,k,r,min, m,n_trai,rank,size );
     
       MPI_Allreduce(&par_e,eb,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
     
	e=*eb;

      if(rank<residuo_trai)
	q_trai=q_trai+1;

     
       if((e<0.5)&&(e!=0.0)){     
	for(j=0;j<q_trai;j++){
	  if(k[j]==1)
	     w[j]=(w[j])/(2.0*e);
	  else
	    w[j]=(w[j])/(2.0*(1.0-e));
	}
      }
      else{ 
	for(j=0;j<tol;j++){

	  X=training_par(data,n_trai,m,p,rank);
	  f=estimated_par( x,X,n_trai,p,m,r,n,rank,size); 
	  ftrai=estimated_par(data,X,n_trai,p,m,n_trai,n,rank,size); 
	  for(i=0;i<q_trai;i++)
	    w[i]=1.0/(double)n_trai;
	  par_e= max_error_par(ftrai,f,w,k_indices,maximos,k,r,min, m ,n_trai,rank,size);
	  MPI_Allreduce(&par_e,eb,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
      
	  if((*eb<0.5)&&(*eb!=0)){
	    break;
	  }
	}

	for(j=0;j<q_trai;j++){
	  if(k[j]==1)
	    w[j]=w[j]/(2.0*(*eb));
	  else
	    w[j]=w[j]/(2.0*(1.0-e));
	}
      }
       
       return maximos;
      
}

int boosting(double * x,int n,int n_boos,int m,int p){
    int i,j=0,r=n/10,e=0,err=0,b;
    int *indice=NULL;
    int **M;
    int *k=NULL;
    double eb,tabulate;
    double *beta;
    double *w=NULL;
    double **f;
    double **tes;
    double **X;
    double **trai;
    double **data;
    
    k=calloc(n,sizeof(int));          
    indice=(int *)calloc(n,sizeof(int));
    M=calloc(n_boos,sizeof(int *));   
    beta=calloc(n_boos,sizeof(double));
    w=calloc(n,sizeof(double));
    X=calloc(n,sizeof(double *));     
    tes=calloc(n,sizeof(double *));   
    trai=calloc(n-r+1,sizeof(double *));  
    data=calloc(n-r+1,sizeof(double *));      
   
    indice=randperm(n);  
    for(i=0;i<n;i++){
      X[i]=x+indice[i]*p; 
      k[i]=X[i][p-1]; 
    }

    if(n%10==0){	
      for(i=0;i<10;i++){
	j=j+r;
	tes=test(data,X,w,j-r,r,n-r,p);  
	for(b=0;b<n_boos;b++){
	  trai=trai_weight(data,w,j-r,j,n-r);  
	  M[b]=max_estimado(data,tes,trai,k,w,n,p,m,r,j-r,&eb);
	  beta[b]=eb/(1.0-eb);        
	}
	e=error(M,k,beta,r,n_boos,m,j-r);
	err+=e;	
	
      }
    }
    else {			
      for(i=0;i<n%10;i++){
	j=j+r+1;
	tes=test(data,X,w,j-r-1,r+1,n-r-1,p); 
	for(b=0;b<n_boos;b++){           
	  trai=trai_weight(data,w,j-r-1,j,n-r-1);
	  M[b]=max_estimado(data,tes,trai,k,w,n,p,m,r+1,j-(r+1),&eb);
	  beta[b]=eb/(1.0-eb);
	}
	e=error(M,k,beta,r+1,n_boos,m,j-(r+1));   
	err+=e;                              
	
      }
      for(i=n%10;i<10;i++){
	j=j+r;
	tes=test(data,X,w,j-r,r,n-r,p);
	for(b=0;b<n_boos;b++){
	  trai=trai_weight(data,w,j-r,j,n-r);
	  M[b]=max_estimado(data,tes,trai,k,w,n,p,m,r,j-r,&eb);
	  beta[b]=eb/(1.0-eb);
	} 
	e=error(M,k,beta,r,n_boos,m,j-r);     
	err+=e;				
      }
    }
    return err;
}

int main(int argc, char * argv[]){
  int rank,size;
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &size);
  if(rank==0){	
    FILE *in;      
    int n,m,p,n_boos,parametro[5],promedio;
    int i,j=0,t,r,e=0,err=0,par_e,b;
    int *indice=NULL;
    int **M;
    int *k=NULL;
    double eb,tiempo;
    double *beta;
    double *w=NULL;
    double *x=NULL;
    double **tes;
    double **X;
    double **trai;
    double **data;
    if(argc!=7){
      fprintf(stderr,"Uso de %s\n  nombre_del_archivo #_de_filas  #_de_columnas #_de_clases #_de_muestras_boostrapeadas #promedio \n",argv[0]);
      exit(1);
    }
    tiempo= MPI_Wtime();
    srand(time(0x0));   
    n=atoi(argv[2]);    
    p=atoi(argv[3]);    
    m=atoi(argv[4]);    
    n_boos=atoi(argv[5]);    
    promedio=atoi(argv[6]);
    parametro[0]=n; 
    parametro[1]=p;
    parametro[2]=m;
    parametro[3]=n_boos;
    parametro[4]=promedio;
    MPI_Bcast(parametro,5,MPI_INT,0,MPI_COMM_WORLD);
    r=n/10;
    k=calloc(n,sizeof(int));          
    indice=(int *)calloc(n,sizeof(int));
    M=calloc(n_boos,sizeof(int *));   
    beta=calloc(n_boos,sizeof(double));
    w=calloc(n,sizeof(double));
    X=calloc(n,sizeof(double *)); 
    x=calloc(n*p,sizeof(double)); 
    tes=calloc(n,sizeof(double *));   
    trai=calloc(n-r+1,sizeof(double *));  
    data=calloc(n-r+1,sizeof(double *));      
    in=fopen(argv[1],"r");        
    if(in==NULL){                 
      printf("problemas con el archivo de salida");
      exit(1);
    }
    while (!feof(in)){       
      fscanf(in,"%lf",&x[j]); 
      j++;
    }
    fclose(in); 
    if(size==1){
      for(i=0;i<promedio;i++){
	e=boosting(x,n,n_boos,m,p);   
	err+=e;
      }
      tiempo=MPI_Wtime()-tiempo;
      printf("\nPROGRAMA %s DATA %s PROCRS  %d  N_BOOST %d ERRORES %lf tiempo paralelo es %.3lf \n: ",argv[0],argv[1],size,n_boos,(double)err/(double)(n*promedio),(double)tiempo/(double)promedio);
      MPI_Finalize(); 
      return 0;
    }
    else{
      j=0;
      r=n/10;
      err=0;
      MPI_Bcast(x,n*p,MPI_DOUBLE,0,MPI_COMM_WORLD);
      for(t=0;t<promedio;t++){
	indice=randperm(n); 
	MPI_Bcast(indice,n,MPI_INT,0,MPI_COMM_WORLD);
	for(i=0;i<n;i++){
	  X[i]=x+indice[i]*p; 
	  k[i]=X[i][p-1]; 
	}	 
	j=0;
	r=n/10;
	if(n%10==0){	
	  for(i=0;i<10;i++){
	    j=j+r;
	    tes=test_par(data,X,w,j-r,r,n-r,p,rank,size);  
	    for(b=0;b<n_boos;b++){
	      trai=trai_weight_par(data,w,j-r,j,n-r,rank,size);  
	      M[b]=max_estimado_par(data,tes,trai,k,w,n,p,m,r,j-r,&eb,rank,size);
	      beta[b]=eb/(1.0-eb);        
	    }
	    par_e=error_par(M,k,beta,n_boos,r,m,j-r,rank,size); 
	    MPI_Reduce(&par_e,&e,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);                          
	     err+=e;	     
          }
	}
	else {	
	  for(i=0;i<n%10;i++){
	    r=n/10;
	    j=j+r+1;
	    tes=test_par(data,X,w,j-r-1,r+1,n-r-1,p,rank,size);              
	    for(b=0;b<n_boos;b++){           
	      trai=trai_weight_par(data,w,j-r-1,j,n-r-1,rank,size);       
	      M[b]=max_estimado_par(data,tes,trai,k,w,n,p,m,r+1,j-(r+1),&eb,rank,size);
	      beta[b]=eb/(1.0-eb);
	    }
	    par_e=error_par(M,k,beta,n_boos,r+1,m,j-(r+1),rank,size);   
	    MPI_Reduce(&par_e,&e,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
	    err+=e;
	  }
	  for(i=n%10;i<10;i++){
	    r=n/10;
	    j=j+r;
	    tes=test_par(data,X,w,j-r,r,n-r,p,rank,size);
	    for(b=0;b<n_boos;b++){
	      trai=trai_weight_par(data,w,j-r,j,n-r,rank,size);
	      M[b]=max_estimado_par(data,tes,trai,k,w,n,p,m,r,j-r,&eb,rank,size);
	      beta[b]=eb/(1.0-eb);
	    } 
	    par_e=error_par(M,k,beta,n_boos,r,m,j-r,rank,size);     
	    MPI_Reduce(&par_e,&e,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
	    err+=e;
	  }
	}
      }
      tiempo=MPI_Wtime()-tiempo;
      printf("\nPROGRAMA %s DATA %s PROCRS  %d  N_BOOST %d PORC DE ERRORES %f tiempo paralelo es %.3lf \n: ",argv[0],argv[1],size,n_boos,(double)err/(double)(n*promedio),(double)tiempo/(double)promedio);
      MPI_Finalize(); 	
    }
    return 0;
  }
  else { /* SLAVE */ 
    int i,j=0,r,t,e=0,err=0,par_e,b,p,cont=0;
    int *indice=NULL;
    int parametro[5];
    int *k=NULL;
    int **M ;
    double eb;
    double *beta;
    double *x=NULL;
    double *w=NULL;
    double **tes;
    double **X;
    double **trai;
    double **data;    
    if(size==1){
      MPI_Finalize(); 
      return 0;
    }
    else{
      MPI_Bcast(parametro,5,MPI_INT,0,MPI_COMM_WORLD);
      r=parametro[0]/10;
      k=calloc(parametro[0],sizeof(int));          
      indice=(int *)calloc(parametro[0],sizeof(int));
      M=calloc(parametro[3],sizeof(int *));
      beta=calloc(parametro[3],sizeof(double));
      w=calloc(parametro[0],sizeof(double));
      x=calloc(parametro[0]*parametro[1],sizeof(double));      
      X=calloc(parametro[0],sizeof(double *));     
      tes=calloc(parametro[0],sizeof(double *));   
      trai=calloc(parametro[0]-r+1,sizeof(double *));  
      data=calloc(parametro[0]-r+1,sizeof(double *));  
      MPI_Bcast(x,parametro[0]*parametro[1],MPI_DOUBLE,0,MPI_COMM_WORLD);
      for(t=0;t<parametro[4];t++){
	MPI_Bcast(indice,parametro[0],MPI_INT,0,MPI_COMM_WORLD);
	for(i=0;i<parametro[0];i++){
	  X[i]=x+indice[i]*parametro[1]; 
	  k[i]=X[i][parametro[1]-1];
	}
	j=0;
	r=parametro[0]/10;
	if(parametro[0]%10==0){
	  for(i=0;i<10;i++){
	    j=j+r;
	    tes=test_par(data,X,w,j-r,r,parametro[0]-r,parametro[1],rank,size);  
	    for(b=0;b<parametro[3];b++){
	      trai=trai_weight_par(data,w,j-r,j,parametro[0]-r,rank,size);  
	      M[b]=max_estimado_par(data,tes,trai,k,w,parametro[0],parametro[1],parametro[2],r,j-r,&eb,rank,size);
	      beta[b]=eb/(1.0-eb);        
	    }
	    par_e=error_par(M,k,beta,parametro[3],r,parametro[2],j-r,rank,size); 
	    MPI_Reduce(&par_e,&e,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
	    err+=e;   
	  }
	}
	else{
	  for(i=0;i<parametro[0]%10;i++){
	    j=j+r+1;
	    tes=test_par(data,X,w,j-r-1,r+1,parametro[0]-r-1,parametro[1],rank,size);              
	    for(b=0;b<parametro[3];b++){           
	      trai=trai_weight_par(data,w,j-r-1,j,parametro[0]-r-1,rank,size);       
	      M[b]=max_estimado_par(data,tes,trai,k,w,parametro[0],parametro[1],parametro[2],r+1,j-(r+1),&eb,rank,size);
	      beta[b]=eb/(1.0-eb);
	    }
	    par_e=error_par(M,k,beta,parametro[3],r+1,parametro[2],j-(r+1),rank,size);
	    MPI_Reduce(&par_e,&e,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD); 
	    err+=e;                              
	  }
	  for(i=parametro[0]%10;i<10;i++){
	    j=j+r;
	    tes=test_par(data,X,w,j-r,r,parametro[0]-r,parametro[1],rank,size);               
	    for(b=0;b<parametro[3];b++){
	      trai=trai_weight_par(data,w,j-r,j,parametro[0]-r,rank,size);
	      M[b]=max_estimado_par(data,tes,trai,k,w,parametro[0],parametro[1],parametro[2],r,j-r,&eb,rank,size);
	      beta[b]=eb/(1.0-eb);
	    } 
	    par_e=error_par(M,k,beta,parametro[3],r,parametro[2],j-r,rank,size);     
	    MPI_Reduce(&par_e,&e,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD); 
	    err+=e;			
	  } 
	}
      }
    }  
    MPI_Finalize(); 
    return 0;
  }
}




