#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#define PI 3.14159


/*funcion que calcula el kernel gaussiano*/
double kernel(double x){
       return 1.0/(exp((x*x)/2.0));   // No se incluye la constante sqrt(2.0*PI) pues da similar resulatado ;
       }

/*funcion que calcula el vector  del ancho de banda para una clase X, se debe ingresar el # de filas de la clase asi como  el # de columanas*/
double * banda(double **X, int n,int p){
       double s,d, Xmean;
       double *h=NULL;
       int i,v;

       h=calloc(n,sizeof(double)); //se asigna memoria dinamica para el vector de salida
       for(v=0;v<p ;v++){
            Xmean=0;
            d=0;
            s=0;
            for(i=0;i<n;i++){
                 s+=X[i][v];
            }
            Xmean =s/n;  //para cada columna se calcula la media por filas
            for(i=0;i<n;i++){
                d+=(X[i][v]-Xmean)*(X[i][v]-Xmean);
            }
            d=sqrt(d/(n-1)); //para cada columna se calcula la desviacion estandar
            h[v]=d*(pow(4.0,1.0/(double)(p+4))/pow((double)((p+2)*n),1.0/(double)(p+4)));

       }
       return h;
       }

 /*funcion que calcula la suma de la productoria de los kernel dado una clase X y un test x, se debe ingresar el vector del *
 *ancho de banda, la iesima fila del test(i), el # de columnas(p) y el # de filas de la clase(m)                           */

 double suma_productoria(double **x,double *h, double **X,int i, int p,int m ){
        double s=0.0,k,t=0.0;
        int j,v;
        for(j=0;j<m ;j++){
                k=1.0;
                for(v=0;v<p;v++){
                     t=kernel((x[i][v]-X[j][v])/h[v]);
                     k=k*t;
                }
              s=s+k;                         //para acumular las productorias y asi obtener la sumatoria.
        }
        return s;
        }

 /*funcion que calcula el estimado del test x dado una clase X , se debe ingresar el vector del ancho de banda (h), el # de filas del test(n) *
  * y el # de columnas(p), el # de filas para una clase del training(m)  y el numero de filas de la matriz original(N)                        */

double * estimado(double **x,double *h,double **X,int n, int p,int m){
       int v,j,i;
       double s=0.0,H=1.0;
       double *f=NULL;
       double pro=(double)m/(double)n;//se calcula la proporcion para el estimado de una clase cualquiera

       f=calloc(n,sizeof(double));//se asigna memoria dinamica para el vector de salida
       for(v=0;v<p;v++)
                   H=H*h[v];   //se calcula la productoria de los elemantos del ancho de  banda
       H=H*(double)m;
       for (i=0;i<n;i++){
              s=suma_productoria(x,h,X,i,p,m);
              f[i]=pro*(s/H);   //se almacena los estimados para cada fila del test
       }
       return f;
       }

 /* funcion que calcula los errores dado una matriz f de dimension(m X n), se debe ingresar el vector de indices de la muestra original(k)  *
 * el # de columnas de f, el # de filas de f, el # de clases(m) y un entero minimo(min) para relizar las comparaciones con el vector k     */
 int error(double **f,int * k_indices ,int n,int m){
      int i,j,e=0,k;
      double max;
       for(j=0;j<n;j++){
             max=0;
             for(i=0;i<m;i++){
                   if(f[i][j]>=max){
                        max=f[i][j];
                        k=i+1;
                    }
             }
             if(k_indices[j]!=k) //se verifica si los indices coenciden o no con la clase, para aumentar el error
                 e++;


       }
       return e;
       }
/* funcion que sirve para imprimir matrices de tipo double en la pantalla, se ingresar la matriz con sus respectivas filas y columnas*/

void print(double **X,int n,int m){
       int i,j;
        for(i=0;i<n;i++){
                for(j=0;j<m;j++)
                      printf("%e\t",X[i][j]);
                 printf("\n");
           }

       }

/*funcion que calcula la matriz de estimados, se le debe ingresar la matriz test(x), matriz training(X) , # de filas del training(n),  *
 * # de columnas(p), # de clases(m), # de filas del test(r), y el # de filas de la matriz original (N)                                 */

int kernel_mul(double *x,int n, int p,int m){
       double **f;
       double ** temp;
       double **X;
       double *h=NULL;
       int i,j,cont,err;
       int *k=NULL;
       k=calloc(n,sizeof(int));
       f=calloc(m,sizeof(double *)); //se asigna memoria dinamica para la matriz de estimados
       temp=calloc(n,sizeof(double *)); //se asigna memoria dinamica para la matriz de clases
       X=calloc(n,sizeof(double *)); //se asigna memoria dinamica para la matriz de clases
       h=calloc(p,sizeof(double));   //se asigna memoria dinamica para el vector del ancho de banda

       for(i=0;i<n;i++){
          X[i]=x+i*p;
          k[i]=X[i][p-1];
       }

       for(j=0;j<m;j++){
            cont=0;
            for(i=0;i<n;i++)
                    if(X[i][p-1]==(double)(j+1)){
                       temp[cont++]= X[i];            //se almacena en temp las filas de la matriz training correspondientes a la clase j
                    }

            h=banda(temp,cont,p-1);                    //se calcula el vector del ancho de banda para una clase dada
            f[j] = estimado(X,h,temp,n,p-1,cont);   //para cada clase j se almacena en la matriz f los estimados para una clase dada

      }

      err=error(f,k,n,m);
      return err;
}


int main(int argc , char *argv[]){

    FILE *in;     //se declara un apuntador de tipo FILE
    double *x=NULL;   //se declara apuntador de tipo double para leer los datos del archivo
    int n,m,p,j=0;
    int err;
    long int passed;
    clock_t  now, later;   //se define dos variables te tipo clock_t para calcular el tiempo

    now = clock();    //se toma el tiempo al inicio del programa.
    n=atoi(argv[2]);  //se convierte el segundo argumento a tipo entero
    p=atoi(argv[3]);   //se convierte el tercer argumento a tipo entero
    m=atoi(argv[4]);   //se convierte el cuarto argumento a tipo entero
    x=malloc(n*p*sizeof(double));//se asiga memoria dinamicamente para el vector donde se va leer
    in=fopen(argv[1],"r");  //se abre el archivo del primer argumento solo para lectura
    if(in==NULL){         //si no se pudo abrir entonces se da un mensaje de error y el programa finaliza
         printf("problemas con el archivo de salida");
         exit(1);
    }
    while (!feof(in)){     //mientras no se acabe los elementos del archivo se seguira leendo
             fscanf(in,"%lf",&x[j]);  //se copia en el vector x los elementos del archivo
             j++;
      }
    fclose(in);    //se cierra el archivo de lectura
    err=kernel_mul(x,n,p,m);
    printf("\n TASA DEL ERROR%f \n",(double)err/(double)n);
    later = clock();   //se toma el tiempo final
    passed =  later - now  ;   //se calcula la diferencia de tiempos para obtener el tiempo real de corrida
    printf("TIEMPO DE EJECUCION EN MICROSEGUNDOS  %lu \n",passed);
    return 0;
}

